{
   Component replicating test.
   Created: 4.11.1995  Mika Nieminen
   Changed: 16.2.1996  Mika Nieminen
             - Changed ReadComponent call in ComponentReplicator to use
               TReader and ReadRootComponent call, because it turn out that
               ReadComponent didn't work with all components. Weird.
}
unit Typetst;

{ If DEBUG_MSG is defined then use debug messages. }
{ Comment out next line to ignore debug messages.  }
{$DEFINE DEBUG_MSG }

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes,
  Controls, Forms, Dialogs, StdCtrls, WinCRT,
  TypInfo, DsgnIntf;

type
  TForm1 = class(TForm)
    Button1: TButton;
    btnRepButton: TButton;
    btnRepListbox: TButton;
    ListBox1: TListBox;
    procedure btnRepButtonClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure Button1MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
  private
    { Private declarations }
    pPreviousButton: TControl;   { Temp pointers to previously }
    pPreviousListbox: TControl;  { duplicated components.      }
  public
    { Public declarations }
  end;

{ Replicate property values of srcComponent and return a new instance }
{ Return NIL if error or pointer to new component if OK }
function ComponentReplicator(srcComponent: TComponent; owner: TComponent): TComponent;

{ Print debug message }
procedure DebugMsg(sMsg: string);

var
  Form1: TForm1;

implementation

{$R *.DFM}

procedure TForm1.FormCreate(Sender: TObject);
begin
	InitWinCRT; { Init CRT window, because it's used to pring debug messages }
  pPreviousButton  := nil;
  pPreviousListBox := nil;
end;


procedure TForm1.btnRepButtonClick(Sender: TObject);
var
  iLeftPos:   integer;    { Left position for the new component }
  oldControl: TComponent;
	newControl: TComponent;
begin
  iLeftPos := 140;
  oldControl := nil;
  newControl := nil;

  { Are we replication a button or are we not? Now, that's a question. }
  if Sender = btnRepButton then oldControl := Button1
  else if Sender = btnRepListbox then oldControl := ListBox1;

  { Replicate property values and event handlers of oldControl to newControl }
  newControl := ComponentReplicator(oldControl, self);

  if (newControl <> nil) AND (newControl is TControl) then begin
     { Make sure newControl won't overflap with the previous one, because }
     { replicator has naturally copied Left and Top properties       }
     if Sender = btnRepButton then begin
        if pPreviousButton <> nil then
           iLeftPos := pPreviousButton.Left + pPreviousButton.Width + 10;
        pPreviousButton := TControl(newControl);
     end else if Sender = btnRepListbox then begin
        if pPreviousListBox <> nil then
           iLeftPos := pPreviousListbox.Left + pPreviousListbox.Width + 10;
        pPreviousListbox := TControl(newControl);
        (newControl as TListBox).Items.Add(IntToStr(iLeftPos));
     end;

     TControl(newControl).Left   := iLeftPos;
     TControl(newControl).Parent := self; { newButton becomes visible at this point }
  end;
end;

procedure TForm1.Button1Click(Sender: TObject);
begin
	DebugMsg((Sender as TComponent).Name + ' was clicked.');
end;

procedure TForm1.Button1MouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
	DebugMsg('Do not tickle me, said ' + (Sender as TComponent).Name);
end;

{----------------------------------------------------------------------------
  Print debug message.
}
procedure DebugMsg(sMsg: string);
begin
{$IFDEF DEBUG_MSG}
  writeln(sMsg);
{$ENDIF}
end;

{----------------------------------------------------------------------------
	Replicate property and event values of srcComponent to a new instance.
  In: srcComponent = Source component (replicated component)
      owner        = Owner of the new component
  Out: NIL = Error, otherwise pointer to a new component
}
function ComponentReplicator(srcComponent: TComponent; owner: TComponent): TComponent;
var
  memStream: TMemoryStream; { Component buffer }
  reader:    TReader;       { Component reader }
  idx: 		 Integer;	{ Running counter           }
	iCount: 	 Integer;	{ Number of properties in a list }
  propList: PPropList;	{ List of property-objects       }
  iSize: 	 Integer;	{ Size of propList in bytes      }
  methodProp: TMethod;	{ Value of method-property       }
  sMethodName: String;	{ Temp string for method names }
begin
	Result   := nil;
	propList := nil;
  reader   := nil;
  memStream:= nil;

  { Make sure srcComponent is valid instance }
  if srcComponent = nil then Exit;

 try

  { If caller didn't create trgComponent, we have to create it now }
{  if trgComponent = nil then
     trgComponent := TComponentClass(srcComponent.ClassType).Create(srcComponent.Owner);
}

  { Make sure components are the same type }
{  if(srcComponent.ClassType <> trgComponent.ClassType) then begin
		DebugMsg(srcComponent.ClassName + ' is not a type of ' + trgComponent.ClassName);
     Exit;
  end;
	DebugMsg('');
}

  { Write ordinal property values of srcComponent into a stream and }
  { read them back to trgComponent. }
  memStream := TMemoryStream.Create;
  memStream.WriteComponent(srcComponent);
  memStream.Position := 0;
  { memStream.ReadComponent(trgComponent); } { Bug! Doesn't work with all components }

  { This is required to get ReadRootComponent-method to work properly. }
  { It doesn't work with unregistered classes. }
  RegisterClasses([TComponentClass(srcComponent.ClassType)]);

  { Setup reader object to memStrean and read component data. }
  { Note! Event the component Name property is duplicated, but }
  { it doesn't matter as long we don't try to use owner.InsertComponent-method }
  reader        := TReader.Create(memStream, 256);
  reader.Parent := owner;
  { Fix! ReadRootComponent flashes the srcComponent for some reasong ?!?! } 
  Result        := reader.ReadRootComponent(nil);

  { Get number of properties (ordinal + event properties) }
  iCount := GetTypeData(srcComponent.ClassInfo)^.PropCount;
  DebugMsg('Total property count is ' + IntToStr(iCount));

	{ Get the number of method-properties }
	iCount := GetPropList(srcComponent.ClassInfo, [tkMethod], nil);
	DebugMsg('Type of ' + srcComponent.ClassName + '(' + srcComponent.Name + ')' +
      		' has ' + IntToStr(iCount) + ' method properties.');

  { Allocate size for propList and fill it with method properties }
  iSize := iCount * SizeOf(Pointer);
  GetMem(propList, iSize);
	GetPropList(srcComponent.ClassInfo, [tkMethod], propList);

  { Browse through all method properties }
	for idx := 0 to iCount - 1 do begin
     { Get the value of method-property from srcComponent and }
   	{ set the same value to trgComponent. }
   	methodProp := GetMethodProp(srcComponent, propList^[idx]);
		SetMethodProp(Result, propList^[idx], methodProp);

     { If event handler was defined then store its name to tmp string }
     if(methodProp.Code <> nil) then
       	sMethodName := srcComponent.Owner.MethodName(methodProp.Code)
     else
        sMethodName := '';   { No method defined }

     { Print the name of a property and possible event handler method }
		DebugMsg(propList^[idx]^.Name + '  ' + sMethodName);
  end;

 finally
   { Free propList and stream allocations }
  if(propList <> nil) then FreeMem(propList, iSize);
  reader.Free;
  memStream.Free;
 end;
end;

end.
